#!/usr/bin/env python3

from moodlews.service import MoodleClient

import welearnbot.action_handlers as handler
from welearnbot import resolvers
from welearnbot.constants import BASEURL, LINK_CACHE
from welearnbot.parser import setup_parser

import errno
import os
import sys


def main():
    # Get command line options
    parser = setup_parser()
    args = parser.parse_args()

    action = resolvers.resolve_action_mode(args)

    config = resolvers.get_config()
    username, password = resolvers.get_credentials(config)

    # Login to WeLearn with supplied credentials
    moodle = MoodleClient(BASEURL)
    token = moodle.authenticate(username, password)
    if not token:
        print("Invalid credentials!")
        sys.exit(errno.EACCES)

    # Select all courses from config if `ALL` keyword is used
    if "ALL" in map(str.upper, args.courses):
        if action == "submissions":
            args.courses = list(resolvers.resolve_submission_details(config).keys())
        else:
            args.courses = resolvers.get_all_courses(config)

    ignore_types = resolvers.resolve_ignore_types(config, args)

    prefix_path = resolvers.resolve_prefix_path(config, args)

    # Store cache file paths
    link_cache_filepath = os.path.join(prefix_path, LINK_CACHE)

    common_args = (
        args,
        config,
        moodle,
        ignore_types,
        prefix_path,
        link_cache_filepath,
        token,
    )
    # Action picker
    if action == "whoami":
        handler.handle_whoami(moodle)

    elif action == "courses":
        handler.handle_courses(moodle)

    elif action == "assignments":
        handler.handle_assignments(*common_args)

    elif action == "submissions":
        handler.handle_submissions(*common_args)

    elif action == "urls":
        handler.handle_urls(args, moodle)

    elif action == "files":
        handler.handle_files(*common_args)


if __name__ == "__main__":
    main()



from argparse import ArgumentParser, RawTextHelpFormatter


def setup_parser() -> ArgumentParser:
    """Setups an argument parser"""
    parser = ArgumentParser(
        description="A command line client for interacting with WeLearn.",
        formatter_class=RawTextHelpFormatter,
    )

    parser.add_argument(
        "action",
        nargs=1,
        help="choose from\n\
    files       - downloads files/resources\n\
    assignments - lists assignments, downloads attachments\n\
    submissions - lists submission count, downloads attachments\n\
    urls        - lists urls\n\
    courses     - lists enrolled courses\n\
    whoami      - shows the user's name and exits\n\
    Abbreviations such as any one of 'f', 'a', 's', 'u', 'c', 'w' are supported.",
    )
    parser.add_argument(
        "courses",
        nargs="*",
        help="IDs of the courses to download files from. The word ALL selects all courses \n\
    from [submissions] section in .welearnrc or welearn.ini for 'submissions' action\n\
    from the [courses] section in .welearnrc or welearn.ini for all other action",
    )
    parser.add_argument("--version", action="version", version="1.3.0")
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="show verbose warnings/errors",
    )
    parser.add_argument(
        "-d",
        "--dueassignments",
        action="store_true",
        help="show only due assignments with the 'assignments' action",
    )
    parser.add_argument(
        "-c",
        "--gcalendar",
        action="store_true",
        help="add due assignments to Google Calendar with the 'assignments' action",
    )
    parser.add_argument(
        "-i",
        "--ignoretypes",
        nargs="*",
        help="ignores the specified extensions when downloading, overrides .welearnrc",
    )
    parser.add_argument(
        "-r",
        "--rolls",
        nargs="*",
        help="roll numbers for which you want to download the submissions using the 'submissions' action",
    )
    parser.add_argument(
        "-p",
        "--pathprefix",
        nargs=1,
        help="save the downloads to a custom path, overrides .welearnrc",
    )
    parser.add_argument(
        "-f",
        "--forcedownload",
        action="store_true",
        help="force download files even if already downloaded/ignored",
    )
    parser.add_argument(
        "-u",
        "--update-course-cache",
        action="store_true",
        help="update course cache. Use this class when you change [submissions] section of config",
    )
    parser.add_argument(
        "-m",
        "--missingdownload",
        action="store_true",
        help="re-download those files which were downloaded earlier but deleted/moved from their location",
    )
    return parser



from moodlews.service import MoodleClient, ServerFunctions
from welearnbot.utils import get_rolls

from argparse import Namespace
from configparser import RawConfigParser
from typing import Tuple, List

import errno
import getpass
import os
import sys


def resolve_action_mode(args: Namespace) -> str:
    # Get the mode
    action = ""
    if "files".startswith(args.action[0]):
        action = "files"
    elif "assignments".startswith(args.action[0]):
        action = "assignments"
    elif "submissions".startswith(args.action[0]):
        action = "submissions"
    elif "urls".startswith(args.action[0]):
        action = "urls"
    elif "courses".startswith(args.action[0]):
        action = "courses"
    elif "whoami".startswith(args.action[0]):
        action = "whoami"
    else:
        print("Invalid action! Use the -h flag for usage.")
        sys.exit(errno.EPERM)

    if args.dueassignments and action != "assignments":
        print(
            "Can only use --dueassignments with 'assignments' action! Use the -h flag for usage."
        )
        sys.exit(errno.EPERM)
    if args.gcalendar and action != "assignments":
        print(
            "Can only use --gcalendar with 'assignments' action! Use the -h flag for usage."
        )
        sys.exit(errno.EPERM)
    if args.rolls and action != "submissions":
        print(
            "Can only use --rolls with 'submission' action! Use the -h flag for usage."
        )
        sys.exit(errno.EPERM)
    return action


def get_config() -> RawConfigParser:
    """Read the .welearnrc file from the home directory, and extract username and password"""
    if sys.platform == "linux" or sys.platform == "linux2":
        configfile = os.path.expanduser("~/.welearnrc")
    elif sys.platform == "darwin":
        configfile = os.path.expanduser("~/.welearnrc")
    elif sys.platform == "win32":
        configfile = os.path.expanduser("~/welearn.ini")

    config = RawConfigParser(allow_no_value=True)
    config.read(configfile)

    return config


def get_credentials(config: RawConfigParser) -> Tuple[str, str]:
    username = ""
    password = ""

    try:
        username = config["auth"]["username"]
    except KeyError:
        username = input("Username : ")
    try:
        password = config["auth"]["password"]
    except KeyError:
        password = getpass.getpass("Password : ", stream=None)

    return username, password


def get_userid(moodle: MoodleClient):
    site_info = moodle.server(ServerFunctions.SITE_INFO)
    return site_info["userid"]


def get_all_courses(config: RawConfigParser) -> List[str]:
    """Also extract the list of `ALL` courses"""
    try:
        all_courses = list(config["courses"].keys())
    except KeyError:
        all_courses = []
    all_courses = map(str.strip, all_courses)
    all_courses = map(str.upper, all_courses)
    all_courses = list(all_courses)
    return all_courses


def resolve_submission_details(config: RawConfigParser,) -> dict[str, List[str]]:
    try:
        courses = dict(config["submissions"])
    except KeyError:
        print("Invalid configuration!")
        print("There is no field '[submissions]' in your config file")
        sys.exit(errno.ENODATA)

    submission_courses = {}
    for (key, val) in courses.items():
        submission_courses[key.upper()] = get_rolls(val)
    return submission_courses


def resolve_ignore_types(config: RawConfigParser, args: Namespace) -> List[str]:
    # Read ignore types from config
    ignore_types = []
    try:
        ignores = config["files"]["ignore"]
        ignore_types = ignores.split(",")
    except KeyError:
        ignore_types = []

    # Override config with options
    if args.ignoretypes:
        ignore_types = args.ignoretypes

    # Override ignore with force
    if args.forcedownload:
        ignore_types = []

    ignore_types = map(str.strip, ignore_types)
    ignore_types = map(str.upper, ignore_types)
    ignore_types = list(ignore_types)

    return ignore_types


def resolve_prefix_path(config: RawConfigParser, args: Namespace) -> str:
    # Read pathprefix from config
    try:
        prefix_path = os.path.expanduser(config["files"]["pathprefix"])
        prefix_path = os.path.abspath(prefix_path)
        if not os.path.isdir(prefix_path):
            print(
                prefix_path,
                "does not exist! Please create an empty directory ",
                prefix_path,
            )
            sys.exit(errno.ENOTDIR)
    except KeyError:
        prefix_path = ""

    # Override pathprefix config if -p flag is used
    if args.pathprefix:
        prefix_path = os.path.expanduser(args.pathprefix[0])
        prefix_path = os.path.abspath(prefix_path)
        if not os.path.isdir(prefix_path):
            print(prefix_path, "does not exist!")
            sys.exit(errno.ENOTDIR)

    return prefix_path


def get_courses_by_id(moodle: MoodleClient, args: Namespace):
    # Get a list of all courses
    courses = moodle.server(ServerFunctions.ALL_COURSES)

    # Create a dictionary of course ids versus course names
    course_ids = dict()
    for course in courses["courses"]:
        course_name = course["shortname"]
        if course_name in args.courses:
            course_ids[course["id"]] = course_name
    return course_ids



from welearnbot.constants import EVENT_CACHE

from typing import Any, Tuple

from configparser import RawConfigParser
from datetime import datetime, timedelta
from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials

import errno
import os
import sys

from welearnbot.utils import create_event, read_cache, write_cache


def setup_gcal(config: RawConfigParser) -> Tuple[str, Any]:
    """Handle Google Calender

    Returns
    -------
    gcal_calendar_id: str
    service: Any
    """
    try:
        OAUTH_CLIENT_ID = config["gcal"]["client_id"]
        OAUTH_CLIENT_SECRET = config["gcal"]["client_secret"]

        gcal_client_config = {
            "installed": {
                "client_id": OAUTH_CLIENT_ID,
                "client_secret": OAUTH_CLIENT_SECRET,
                "redirect_uris": ["http://localhost", "urn:ietf:wg:oauth:2.0:oob"],
                "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                "token_uri": "https://accounts.google.com/o/oauth2/token",
            }
        }
    except KeyError:
        print("Invalid configuration!")
        sys.exit(errno.ENODATA)

    try:
        gcal_calendar_id = config["gcal"]["calendar_id"]
    except KeyError:
        gcal_calendar_id = "primary"

    # Connect to the Google Calendar API
    SCOPES = ["https://www.googleapis.com/auth/calendar.events"]
    creds = None
    gcal_token_path = os.path.expanduser("~/.gcal_token")
    if os.path.exists(gcal_token_path):
        creds = Credentials.from_authorized_user_file(gcal_token_path, SCOPES)
    # If there are no valid credentials, login
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_config(gcal_client_config, SCOPES)
            creds = flow.run_local_server(port=0)
        # Save the credentials for the next run
        with open(gcal_token_path, "w") as gcal_token:
            gcal_token.write(creds.to_json())
    service = build("calendar", "v3", credentials=creds)

    return gcal_calendar_id, service


def publish_gcal_event(
    config: RawConfigParser,
    duedate: datetime,
    course_name: str,
    name: str,
    assignment_id: int,
    detail: str,
) -> None:
    event_cache_filepath = os.path.expanduser(EVENT_CACHE)
    event_cache = read_cache(event_cache_filepath)

    # Put deadline at the *end* of the event
    startdate = duedate - timedelta(hours=1)
    start_time = startdate.isoformat()
    end_time = duedate.isoformat()
    event_name = f"{course_name} - {name}"

    gcal_calendar_id, service = setup_gcal(config)

    assignment_id = str(assignment_id)
    if assignment_id not in event_cache:
        # Create and push a new event
        event = create_event(event_name, detail, start_time, end_time, False)
        added_event = (
            service.events().insert(calendarId=gcal_calendar_id, body=event).execute()
        )
        event_id = added_event["id"]
        event_cache[assignment_id] = event_id
        print(f"        Added event to calendar.")
    else:
        # Update event if necessary
        event = (
            service.events()
            .get(calendarId=gcal_calendar_id, eventId=event_cache[assignment_id],)
            .execute()
        )
        if event["start"]["dateTime"] != (start_time + "+05:30"):
            event["start"]["dateTime"] = start_time
            event["end"]["dateTime"] = end_time
            updated_event = (
                service.events()
                .update(calendarId=gcal_calendar_id, eventId=event["id"], body=event,)
                .execute()
            )
            event_cache[assignment_id] = updated_event["id"]
            print(f"        Updated event in calendar.")
    write_cache(event_cache_filepath, event_cache)



BASEURL = "https://welearn.iiserkol.ac.in"
LINK_CACHE = ".link_cache"
COURSE_CACHE = ".course_cache"
EVENT_CACHE = "~/.welearn_event_cache"
ARCHIVE_TYPES = [
    "7Z",
    "GZ",
    "GZIP",
    "TAR",
    "TGZ",
    "ZIP",
]



from moodlews.service import MoodleClient, ServerFunctions
from welearnbot.constants import ARCHIVE_TYPES

from argparse import Namespace
from typing import Any, Dict, List, Tuple

import json
import os
import mimetypes
import zipfile


def read_cache(filepath: str) -> dict:
    """Read from a cache file"""
    cache = dict()
    if os.path.exists(filepath):
        with open(filepath) as cache_file:
            cache = json.load(cache_file)
    return cache


def write_cache(filepath: str, cache: dict) -> None:
    """Update cache file"""
    with open(filepath, "w") as cache_file:
        json.dump(cache, cache_file)


def get_courses_cache(
    moodle: MoodleClient, course_cache_filepath: str, userid: str, update: bool = False,
):
    cache = read_cache(course_cache_filepath)
    if update or not cache:
        courses = moodle.server(ServerFunctions.USER_COURSES, {"userid": userid})
        for course in courses:
            courseid = course["shortname"]
            course_data = {
                "id": course["id"],
                "courseid": course["shortname"],
                "name": course["fullname"],
                "participants": {},
            }
            participants = moodle.server(
                ServerFunctions.COURSE_USERS, {"courseid": course["id"]}
            )
            for participant in participants:
                try:
                    print(participant["idnumber"])
                    course_data["participants"][participant["idnumber"]] = {
                        "id": participant["id"],
                        "name": participant["fullname"],
                    }
                except KeyError:
                    # skip caching details of participants whose information is not available.
                    # This mostly happens for instructors
                    continue
            cache[courseid] = course_data
    with open(course_cache_filepath, "w") as f:
        json.dump(cache, f)
    return cache


def fetch_assignments(moodle, courseid) -> List[Any]:
    assignments = moodle.server(
        ServerFunctions.ASSIGNMENTS, {"courseids[0]": courseid}  # cache[course]["id"]}
    )
    assignments = assignments["courses"][0]["assignments"]
    return [
        {
            "id": a["id"],
            "name": a["name"],
            "fileurl": a["introattachments"][0]["fileurl"]
            if a["introattachments"]
            else "",
            "duedate": a["duedate"],
            "opendate": a["allowsubmissionsfromdate"],
        }
        for a in assignments
    ]


def create_event(
    name: str, description: str, start: str, end: str, reminders: bool = True
) -> Dict[str, Any]:
    """Format and create a calendar event"""
    newevent = {
        "summary": name,
        "location": "",
        "description": description,
        "start": {"dateTime": start, "timeZone": "Asia/Kolkata",},
        "end": {"dateTime": end, "timeZone": "Asia/Kolkata",},
        "reminders": {
            "useDefault": reminders,
            "overrides": [{"method": "popup", "minutes": 10},],
        },
    }
    return newevent


def download_resource(
    args: Namespace,
    moodle: MoodleClient,
    ignore_types: List[str],
    resource: Any,
    prefix: str,
    course: str,
    cache: dict,
    token: str,
    subfolders: List[str] = [],
    indent: int = 0,
    extract: bool = True,
) -> Tuple[str, str]:
    """Helper function to retrieve a file/resource from the server"""
    filename = resource["filename"]
    course_dir = os.path.join(prefix, course, *subfolders)
    fileurl = resource["fileurl"]
    _, extension = os.path.splitext(filename)
    extension = str.upper(extension[1:])
    if extension == "":
        # Missing extension - guess on the basis of the mimetype
        extension = mimetypes.guess_extension(resource["mimetype"])
        filename += extension
        extension = extension[1:]
    filepath = os.path.join(course_dir, filename)
    short_filepath = os.path.join(course, *subfolders, filename)
    timemodified = int(resource["timemodified"])

    # Only download if forced, or not already downloaded
    if not args.forcedownload and fileurl in cache:
        cache_time = int(cache[fileurl])
        # Check where the latest version of the file is in cache
        if timemodified == cache_time:
            if os.path.exists(filepath):
                return "EXISTS", short_filepath
            if not args.missingdownload and not os.path.exists(filepath):
                return "MISSING", short_filepath

    # Ignore files with specified extensions
    if extension in ignore_types:
        return "IGNORE", short_filepath

    # Create the course folder if not already existing
    if not os.path.exists(course_dir):
        os.makedirs(course_dir)

    # Download the file and write to the folder
    print(
        " " * indent + "Downloading " + short_filepath, end="", flush=True,
    )
    response = moodle.response(fileurl, token=token)
    with open(filepath, "wb") as download:
        download.write(response.content)

    # TODO: add option whether to extract or not in the config and flag
    if extension in ARCHIVE_TYPES and extract:
        with zipfile.ZipFile(filepath, "r") as zip_ref:
            zip_ref.extractall(course_dir)
            print(" ... EXTRACTING", end="", flush=True)

    print(" ... DONE")

    # Add the file url to the cache
    cache[fileurl] = timemodified
    return "DOWNLOADED", short_filepath


def show_file_statuses(file_statuses, verbose=False) -> None:
    """Helper function to print ignored, missing files"""
    ignored = []
    missing = []
    downloaded = []
    for status, short_filepath in file_statuses:
        if status == "IGNORE":
            ignored.append(short_filepath)
        elif status == "MISSING":
            missing.append(short_filepath)
        elif status == "DOWNLOADED":
            downloaded.append(short_filepath)

    if len(ignored) > 0:
        if len(downloaded) > 0:
            print()
        if verbose:
            print("The following files have been ignored.")
            for short_filepath in ignored:
                print("    " + short_filepath)
        else:
            if len(ignored) == 1:
                print("1 file has been ignored, use --verbose for more info")
            else:
                print(
                    "{} files have been ignored, use --verbose for more info".format(
                        len(ignored)
                    )
                )

    if len(missing) > 0:
        if len(ignored) > 0 or len(downloaded) > 0:
            print()
        if verbose:
            print(
                "The following files are missing, use --missingdownload to download them."
            )
            for short_filepath in missing:
                print("    " + short_filepath)
        else:
            if len(missing) == 1:
                print("1 file is missing, use --verbose for more info")
            else:
                print(
                    "{} files are missing, use --verbose for more info".format(
                        len(missing)
                    )
                )


def get_rolls(roll_string: str) -> List[str]:
    roll_string = roll_string.strip().upper()
    if "ALL" in roll_string:
        return ["ALL"]
    raw_rolls = roll_string.split(",")
    rolls = []
    for i, roll in enumerate(raw_rolls):
        if roll == "...":
            rolls.extend(expand_dots(raw_rolls[i - 1], raw_rolls[i + 1]))
        else:
            rolls.append(roll)
    return rolls


def expand_dots(start: str, end: str):
    """
    Gives roll nos. between `start` and `end`
    """
    batch = start[:4]
    if batch != end[:4]:
        raise ValueError("Batch id before and after '...' does not match")
    start_roll = int(start[4:])
    end_roll = int(end[4:])
    return [f"{batch}{roll:03}" for roll in range(start_roll + 1, end_roll)]



from argparse import Namespace
from configparser import RawConfigParser
import os
from time import time
from typing import List
from bs4 import BeautifulSoup as bs
from datetime import datetime

from moodlews.service import MoodleClient, ServerFunctions
from welearnbot import resolvers, utils
from welearnbot.constants import COURSE_CACHE
from welearnbot.gcal import publish_gcal_event


def handle_whoami(moodle: MoodleClient) -> None:
    info = moodle.server(ServerFunctions.SITE_INFO)
    print(info["fullname"])


def handle_courses(moodle: MoodleClient) -> None:
    # Get core user information
    info = moodle.server(ServerFunctions.SITE_INFO)
    userid = info["userid"]

    # Get enrolled courses information
    courses = moodle.server(ServerFunctions.USER_COURSES, {"userid": userid})
    for course in courses:
        course_name = course["fullname"]
        star = " "
        if course["isfavourite"]:
            star = "*"
        print(f" {star} {course_name}")


def handle_assignments(
    args: Namespace,
    config: RawConfigParser,
    moodle: MoodleClient,
    ignore_types: List[str],
    prefix_path: str,
    link_cache_filepath: str,
    token: str,
) -> None:
    link_cache = utils.read_cache(link_cache_filepath)
    # Get assignment data from server
    assignments = moodle.server(ServerFunctions.ASSIGNMENTS)

    file_statuses = []
    # Assignments are grouped by course
    for course in assignments["courses"]:
        course_name = course["shortname"]
        # Ignore unspecified courses
        if course_name not in args.courses:
            continue
        no_assignments = True
        for assignment in course["assignments"]:
            # Get the assignment name, details, due date, and relative due date
            assignment_id = assignment["id"]
            name = assignment["name"]
            duedate = datetime.fromtimestamp(int(assignment["duedate"]))
            due_str = duedate.strftime("%a %d %b, %Y, %H:%M:%S")
            duedelta = duedate - datetime.now()
            # Calculate whether the due date is in the future
            due = duedelta.total_seconds() > 0
            if args.dueassignments and not due:
                continue
            no_assignments = False
            if not no_assignments:
                print(course_name)
            # Show assignment details
            duedelta_str = (
                f"{abs(duedelta.days)} days, {duedelta.seconds // 3600} hours"
            )
            detail = bs(assignment["intro"], "html.parser").text
            print(f"    {name} - {detail}")
            for attachment in assignment["introattachments"]:
                print(f"        Attachment     : {attachment['filename']}")
                file_statuses.append(
                    utils.download_resource(
                        args,
                        moodle,
                        ignore_types,
                        attachment,
                        prefix_path,
                        course_name,
                        link_cache,
                        token,
                        indent=8,
                    )
                )
            if due:
                print(f"        Due on         : {due_str}")
                print(f"        Time remaining : {duedelta_str}")
            else:
                print(f"        Due on         : {due_str} ({duedelta_str} ago)")

            # Get submission details
            submission = moodle.server(
                ServerFunctions.ASSIGNMENT_STATUS, {"assignid": assignment_id}
            )
            submission_made = False
            try:
                for plugin in submission["lastattempt"]["submission"]["plugins"]:
                    if plugin["name"] == "File submissions":
                        for filearea in plugin["fileareas"]:
                            if filearea["area"] == "submission_files":
                                for submitted_file in filearea["files"]:
                                    submission_made = True
                                    filename = submitted_file["filename"]
                                    submission_date = datetime.fromtimestamp(
                                        int(submitted_file["timemodified"])
                                    )
                                    submission_date_str = submission_date.strftime(
                                        "%a %d %b, %Y, %H:%M:%S"
                                    )
                                    print(
                                        f"        Submission     : {filename} ({submission_date_str})"
                                    )
            except KeyError:
                continue
            if not submission_made:
                print(f"        Submission     : NONE")

            # Write event to calendar
            if args.gcalendar and due:
                publish_gcal_event(
                    config, duedate, course_name, name, assignment_id, detail
                )
            print()

    utils.write_cache(link_cache_filepath, link_cache)
    utils.show_file_statuses(file_statuses, verbose=args.verbose)


def handle_submissions(
    args: Namespace,
    config: RawConfigParser,
    moodle: MoodleClient,
    ignore_types: List[str],
    prefix_path: str,
    link_cache_filepath: str,
    token: str,
) -> None:
    userid = resolvers.get_userid(moodle)

    submission_config = resolvers.resolve_submission_details(config)

    course_cache_filepath = os.path.join(prefix_path, COURSE_CACHE)
    courses_cache = utils.get_courses_cache(
        moodle, course_cache_filepath, userid, args.update_course_cache,
    )

    link_cache = utils.read_cache(link_cache_filepath)
    file_statuses = []
    for course in args.courses:
        if course not in courses_cache:
            print(f"{course} is not a valid course id")
            continue
        if args.rolls:
            rolls = utils.get_rolls(",".join(args.rolls))
        else:
            try:
                rolls = submission_config[course]
            except KeyError:
                print(
                    f'Could not resolve roll numbers for {course}. Please add it in your config or use "-r" flag'
                )
                continue
        if "ALL" in rolls:
            rolls = sorted(courses_cache[course]["participants"].keys())

        assignments = utils.fetch_assignments(moodle, courses_cache[course]["id"])
        for assignment in assignments:
            if assignment["duedate"] > time():
                continue
            for roll in rolls:
                submission_data = moodle.server(
                    ServerFunctions.SUBMISSION,
                    {
                        "assignid": assignment["id"],
                        "userid": courses_cache[course]["participants"][roll]["id"],
                    },
                )
                try:
                    file_data = submission_data["lastattempt"]["submission"]["plugins"][
                        0
                    ]["fileareas"][0]["files"]
                except KeyError:
                    continue
                if file_data:
                    file_statuses.append(
                        utils.download_resource(
                            args,
                            moodle,
                            ignore_types,
                            file_data[0],
                            prefix_path,
                            course,
                            link_cache,
                            token,
                            ["submissions", assignment["name"], roll],
                        )
                    )
    utils.write_cache(link_cache_filepath, link_cache)
    utils.show_file_statuses(file_statuses, verbose=args.verbose)


def handle_urls(args: Namespace, moodle: MoodleClient) -> None:
    course_ids = resolvers.get_courses_by_id(moodle, args)

    # Get a list of available urls
    urls = moodle.server(ServerFunctions.URLS)

    # Iterate through all urls, and build a dictionary
    url_list = dict()
    for url in urls["urls"]:
        if url["course"] in course_ids:
            course_name = course_ids[url["course"]]
            if not course_name in url_list:
                url_list[course_name] = []
            url_list[course_name].append(url)

    # Display all urls
    for course_name in args.courses:
        if not course_name in url_list:
            continue
        no_url = True
        for url in url_list[course_name]:
            if no_url:
                print(course_name)
            no_url = False
            url_name = url["name"]
            url_detail = bs(url["intro"], "html.parser").text
            url_link = url["externalurl"]
            print(f"    {url_name} - {url_detail}")
            print(f"        Link : {url_link}")
            print()
        print()


def handle_files(
    args: Namespace,
    config: RawConfigParser,
    moodle: MoodleClient,
    ignore_types: List[str],
    prefix_path: str,
    link_cache_filepath: str,
    token: str,
) -> None:
    link_cache = utils.read_cache(link_cache_filepath)
    course_ids = resolvers.get_courses_by_id(moodle, args)

    file_statuses = []

    # Iterate through each course, and fetch all modules
    for courseid in course_ids:
        course_name = course_ids[courseid]
        page = moodle.server(ServerFunctions.COURSE_CONTENTS, {"courseid": courseid})
        for item in page:
            modules = item.get("modules", [])
            for module in modules:
                modname = module.get("modname", "")
                if modname == "resource":
                    for resource in module["contents"]:
                        file_statuses.append(
                            utils.download_resource(
                                args,
                                moodle,
                                ignore_types,
                                resource,
                                prefix_path,
                                course_name,
                                link_cache,
                                token,
                            )
                        )
                elif modname == "folder":
                    folder_name = module.get("name", "")
                    for resource in module["contents"]:
                        file_statuses.append(
                            utils.download_resource(
                                args,
                                moodle,
                                ignore_types,
                                resource,
                                prefix_path,
                                course_name,
                                link_cache,
                                token,
                                subfolders=[folder_name],
                            )
                        )

    utils.write_cache(link_cache_filepath, link_cache)
    utils.show_file_statuses(file_statuses, verbose=args.verbose)
